//+------------------------------------------------------------------+
//|                                                    USDCAD AI.mq5 |
//|                                        Gamuchirai Zororo Ndawana |
//|                          https://www.mql5.com/en/gamuchiraindawa |
//+------------------------------------------------------------------+
#property copyright "Gamuchirai Zororo Ndawana"
#property link      "https://www.mql5.com/en/gamuchiraindawa"
#property version   "1.00"

//+------------------------------------------------------------------+
//| Resources                                                        |
//+------------------------------------------------------------------+
#resource  "\\Files\\USDCAD XAUUSD XPDUSD M1 Float.onnx" as const uchar onnx_buffer[];

//+------------------------------------------------------------------+
//| Libraries we need                                                |
//+------------------------------------------------------------------+
#include <Trade/Trade.mqh>
CTrade Trade;

//+------------------------------------------------------------------+
//| Global variables                                                 |
//+------------------------------------------------------------------+
long onnx_model;
double mean_values[12],std_values[12];
vectorf model_output = vectorf::Zeros(1);
int model_forecast,state;
double ask,bid;

//+------------------------------------------------------------------+
//| Expert initialization function                                   |
//+------------------------------------------------------------------+
int OnInit()
  {
//--- This function will load our ONNX model
if(!load_onnx())
   {
      return(INIT_FAILED);
   }

//--- This function will load our scaling factors
if(!load_scaling_factors())
   {
      return(INIT_FAILED);
   }

//---
   return(INIT_SUCCEEDED);
  }
//+------------------------------------------------------------------+
//| Expert deinitialization function                                 |
//+------------------------------------------------------------------+
void OnDeinit(const int reason)
  {
//--- Free up the resources we do not need
OnnxRelease(onnx_model);
ExpertRemove();
   
  }
//+------------------------------------------------------------------+
//| Expert tick function                                             |
//+------------------------------------------------------------------+
void OnTick()
  {
//--- Update the market prices
update_market_prices();

//--- Fetch a forecast from our model
model_predict();

//--- Find a trading oppurtunity
if(PositionsTotal() == 0)
   {
      if(model_forecast == -1)
      {
            Trade.Sell(0.2,_Symbol,ask,0,0,"USDCAD AI");
            state = -1;
      }

   else if(model_forecast == 1)
      {
            Trade.Buy(0.2,_Symbol,ask,0,0,"USDCAD AI");
            state = 1;
      }
   }

//--- Check for reversals
if(PositionsTotal() > 0)
   {
      if(state != model_forecast)
         {
            Alert("Reversal detected by our AI system, closing all positions now!");
            Trade.PositionClose(_Symbol);
         }
   }

  }
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//| Model predict                                                    |
//+------------------------------------------------------------------+
void model_predict(void)
   {
 //--- First fetch the market data
      
      vectorf model_input  = { 
                               iOpen("USDCAD",PERIOD_CURRENT,0),iHigh("USDCAD",PERIOD_CURRENT,0),iLow("USDCAD",PERIOD_CURRENT,0),iClose("USDCAD",PERIOD_CURRENT,0), 
                               iOpen("XAUUSD",PERIOD_CURRENT,0),iHigh("XAUUSD",PERIOD_CURRENT,0),iLow("XAUUSD",PERIOD_CURRENT,0),iClose("XAUUSD",PERIOD_CURRENT,0),
                               iOpen("XPDUSD",PERIOD_CURRENT,0),iHigh("XPDUSD",PERIOD_CURRENT,0),iLow("XPDUSD",PERIOD_CURRENT,0),iClose("XPDUSD",PERIOD_CURRENT,0)
                             };
 
 //--- Now standardize and scale the data
      for(int i =0; i < 12; i++)
         {
            model_input[i] = ((model_input[i] - mean_values[i]) / std_values[i]);
         }
 
 //--- Now fetch a prediction from our model
      OnnxRun(onnx_model,ONNX_DEFAULT,model_input,model_output);
 
 //--- Store our model's forecat
 if(model_output[0] > iClose("USDCAD",PERIOD_CURRENT,0))
      {
         model_forecast = 1;
      }
 else if(model_output[0] < iClose("USDCAD",PERIOD_CURRENT,0))
      {
         model_forecast = -1;
      }      
   }

//+------------------------------------------------------------------+
//| Update market prices                                             |
//+------------------------------------------------------------------+
void update_market_prices(void)
   {
      ask = SymbolInfoDouble(_Symbol,SYMBOL_ASK);
      bid = SymbolInfoDouble(_Symbol,SYMBOL_BID);
   }

//+------------------------------------------------------------------+
//| This function will read our scaling factors and store them       |
//+------------------------------------------------------------------+
bool load_scaling_factors(void)
   {
   //--- Read in the file
   string file_name = "usd_cad_xau_xpd_scaling_factors.csv";

//--- Try open the file
   int result = FileOpen(file_name,FILE_READ|FILE_CSV|FILE_ANSI,","); //Strings of ANSI type (one byte symbols). 

//--- Check the result
   if(result != INVALID_HANDLE)
     {
      Print("Opened the file");
      //--- Store the values of the file
      
      int counter = 0;
      string value = "";
      
      while(!FileIsEnding(result) && !IsStopped()) //read the entire csv file to the end 
       {
       
         if (counter > 100) //if you aim to read 10 values set a break point after 10 elements have been read
           break; //stop the reading progress
         
         value = FileReadString(result);
         Print("Trying to read string: ",value," count value: ",counter);
         
         //--- Check where we are
         if((counter >= 14) && (counter < 26))
            {
               mean_values[counter - 14] = (float) value;
            }   
         //--- Check where we are
         if((counter >= 27) && (counter < 39))
            {
               std_values[counter - 27] = (float) value;
            }   
         //--- Reading a new row
         if(FileIsLineEnding(result))
           { 
             Print("row++");
           }
         
         counter++;
       }
      //---Close the file
      ArrayPrint(mean_values);
      ArrayPrint(std_values);
      FileClose(result);
      return(true);
     }
//--- We failed to find the file
else 
   {
      Comment("Failed to find the file containing scaling factors");
      return(false);
   }

   }

//+------------------------------------------------------------------+
//| This function is responsible for loading our ONNX model          |
//+------------------------------------------------------------------+
bool load_onnx(void)
   {
      //--- First we must create the model from the buffer
      onnx_model = OnnxCreateFromBuffer(onnx_buffer,ONNX_DEFAULT);
      
      //--- Now we shall define our I/O shape
      ulong input_shape [] = {1,12};
      ulong output_shape []  = {1,1};
      
      if(!OnnxSetInputShape(onnx_model,0,input_shape))
         {
            Comment("Failed to set ONNX input shape!");
            return(false); 
         }
      
      if(!OnnxSetOutputShape(onnx_model,0,output_shape))
         {
            Comment("Failed to set ONNX output shape!");
            return(false); 
         }
         
      //--- Everything went fine
      return(true);
   }